"""
This script doesn't provide any extra functionality; it just creates a wrapper class for the HEC-RAS controller
that documents important methods, making everything else easier.  Arguments to methods etc are more or less
the same, at most shuffled around a bit.  No actual processing work is done here.  Many methods are renamed
as the original names were somewhat awkward

See https://engineerpaige.com/2018/11/hecras-controller/ for how to open the object browser in Excel.  This
may be helpful for finding useful methods and their arguments.

This module should generally not be directly used from outside the "ras" module.  Other parts of the "ras"
module should wrap all necessary functionality in a more user-friendly manner.
"""

"""
The following HEC-RAS COM methods must therefore have wrappers available:
Current<GeomFile, ProjectFile, SteadyFile, UnSteadyFile>
SetMann
SetMann_LChR
Plan_<GetFilename, SetCurrent>

Mainly just SetMann and SetMann_LChR are needed, as the other stuff can be done through the GUI

Geometry_SetMann(string River, string Reach, string RS, int nMann, Single[] Mann_n, Single[] station, string errmsg)
Geometry_SetMann_LChR(string River, string Reach, string RS, Single MannLOB, Single MannChan, Single MannROB, string errmsg)
"""

from win32com import client


class RasObject(object):
    def __init__(self, rasName = "RAS507.HECRASController", ras=None):
        # By default, initialize a new RAS controller object; otherwise, use the provided one.
        if ras is None:
            self.ras = client.Dispatch(rasName)
        else:
            self.ras = ras

    def ShowRas(self):
        # Show HEC-RAS window
        self.ras.ShowRas()

    def Save(self):
        self.ras.Project_Save()

    def CurrentProject(self):
        return self.ras.Project_Current()

    def OpenProject(self, path):
        # Open a project at the given path
        self.ras.Project_Open(path)

    def GetNodes(self, riv, rch):
        """
        Get information on geometry nodes.
        :param riv: integer river
        :param rch: integer reach
        :return: (riv, rch, (node river stations))
        """
        return self.ras.Geometry_GetNodes(riv, rch)

    def GetRivers(self):
        """
        Get list of rivers.
        :return: (n. rivers, ("river a", "river b", ...))
        """
        return self.ras.Geometry_GetRivers()

    def GetReaches(self, riv):
        """
        Get list of reaches.
        :param riv: integer river number (1 indexed).
        :return: (riv, n. reaches, (reach names, ...))
        """
        return self.ras.Geometry_GetReaches(riv)

    def GetNodeOutput(self, riv, rch, n, updn = None, prof = None):
        """
        Get node output.
        :param riv: int river
        :param rch: int reach
        :param n: int node number (from top)
        :return: (some float (output?), riv, rch, n, updn, prof, nVar)
        """
        return self.ras.Output_NodeOutput(riv, rch, n, updn, prof)

    def GetVelDist(self, riv, rch, n, updn = None, prof = 1):
        """
        Get vel dist.  What is that?
        :param riv: int river
        :param rch: int reach
        :param n: int node number (from top)
        :return: (riv, rch, n, updn, prof[ile?], nv, (left stations), (right stations), (conv percs), (areas),
            (wetted perimeters), (flows), (depths), (velocities)
        """
        return self.ras.Output_VelDist(riv, rch, n, updn, prof)

    def GetVariables(self):
        """
        Get a list of output variables.
        :return: (n. variables, (variable names), (variable descriptions))
        """
        return self.ras.Output_Variables()

    def Compute(self):
        """
        Run the current plan.
        :return: (some bool, nmsg, (messages), blockingmode)
        """
        return self.ras.Compute_CurrentPlan()

    def Complete(self):
        return self.ras.Compute_Complete()

    def SetMannLCR(self, river, reach, rs, left, channel, right):
        # Geometry_SetMann_LChR(string River, string Reach, string RS, Single MannLOB, Single MannChan, Single MannROB, string errmsg)
        """
        Set the Manning's n for left, main channel, right.
        :param river: river name (string)
        :param reach: reach name (string)
        :param rs: river station (string)
        :param left: left bank Manning's n
        :param right: right bank Manning's n
        :param channel: main channel Manning's n
        :return: (arguments, errmsg)
        """
        result = self.ras.Geometry_SetMann_LChR(river, reach, rs, left, channel, right)
        self.ras.Project_Save()
        return result

    def SetMann(self, river, reach, rs, manns):
        # Geometry_SetMann(string River, string Reach, string RS, int nMann, Single[] Mann_n, Single[] station, string errmsg)
        """
        Set all Manning's n at various stations.
        :param river: river name (string)
        :param reach: reach name (string)
        :param rs: river station (string)
        :param manns: Manning's ns
        :param stations: stations (from the left)
        :return: (arguments, errmsg)
        """
        nMann = len(manns)
        result = self.ras.Geometry_SetMann(river, reach, rs, nMann, manns)
        self.ras.Project_Save()
        return result

    def QuitRas(self):
        self.ras.QuitRas()

    def EditSteadyFlow(self):
        self.ras.Edit_SteadyFlowData()

    def SetSteadyFlow(self, river, reach, rs, flows, wait = False):
        # [0] because it starts from flows[1] for some reason
        # May want to consider using PyRASFile for this.  The COM is astonishingly finicky.  Or buggy.
        # I think part of the COM is using 1-based indexing and part of it is using 0-based indexing.
        # Setting the flow throws an index out of range error unless len(flows) = nProfile + 1
        # But saving throws an index out of range error then
        # However, setting n = len(flows) - 1, then setting flows, then setting n = len(flows) still throws
        # an out of range error when saving
        # The problem seems to be changing nProfile, period; whatever it is set to, the graphical editor
        # will always have an error loading flow data.  The problem, presumably, is that nProfile is meant
        # to be read and not written to, even though it does allow writing.  How, then, to set it?
        # One thing that works is asking the user to set it.
        if wait:
            print("Set the number of profiles to %d and hit enter when done" % len(flows))
            self.EditSteadyFlow()
            input()
        self.ras.SteadyFlow_SetFlow(river, reach, rs, [0] + flows)
        # self.ras.SteadyFlow_nProfile = len(flows)  # only works if RAS is not displayed
        self.ras.Project_Save()




"""
Outer wrapper for lower-level functionality.
"""

"""
The following classes and methods are required (note: class names don't need to be as listed here,
but methods and properties do)

rasObj - containing a project with a specified geometry file and plan etc
rasObj.reach(river, reach, geom = None) - returning the relevant reach.  If geom (optional) is specified,
    use that geometry file (specified as an ID, if string, or file number, if integer); otherwise, use the currently
    active geometry, if any, or throw an error if no geometry is active
rasObj.openProject(projectPath) - open the relevant project
rasObj.setPlan(plan) - set the plan to that one (by ID or file number)
rasObj.setGeom(geom) - set the plan geometry file/active geometry file (by ID or file number)
rasObj.computeSteady, computeUnsteady(plan = None) - run the plan (optional) or, if plan is None, the currently active plan
rasObj.exit() - exit HEC-RAS
rasObj.newPlan(planId) - create a new plan with the given ID and make it the active plan
rasObj.set<Steady/Unsteady>Flow(flow) - set the steady/unsteady flow file (by ID or file number)
rasObj.getSimData(river = None, reach = None, rs = None): get the simulation flow data for a given river, reach, and river station.
    If rs is none, return a dictionary of all the river stations' data.  Likewise river and reach (nested dictionaries).
    Simulation flow data should be as a simdata class.

reach - containing a reach in the geometry file
reach.xses - return a list of cross sections

simdata - simulation data for a given cross section
properties: velocity, maxDepth, flow, and other data

xs - a cross section
xs.setAllManning(ns): set all Manning's ns to a list of ns given
xs.setMainChannelManning(n): set the main channel n to the given n
"""

from raspy.ras.wrapper import RasObject

class Ras(object):
    """
    The whole RAS controller.
    """
    def __init__(self, projectPath, rasObject = RasObject()):
        self.ras = rasObject
        if not (projectPath is None):
            self.openProject(projectPath)
        self.rivers = [River(self.ras, river.strip()) for river in self.ras.GetRivers()[1]]

    def openProject(self, path):
        self.ras.OpenProject(path)

    def quit(self):
        self.ras.QuitRas()

    def currentProject(self):
        return self.ras.CurrentProject()

    def save(self):
        self.ras.Save()

    def river(self, river, geom = None):
        return [riv for riv in self.rivers if riv.river == river][0]

    def reach(self, river, reach, geom = None):
        # Geom not implemented yet.
        return self.river(river, geom).reach(reach)

    def computeSteady(self, plan = None):
        self.ras.Compute()

    def computeUnsteady(self, plan = None):
        self.ras.Compute()

    def computeIsComplete(self):
        return self.ras.Complete()

    def getSimData(self, river = None, reach = None, rs = None, prof = 1):
        if river is None:
            return {riv.river: self.getSimData(riv.river, reach, rs) for riv in self.rivers}
        elif reach is None:
            riv = self.river(river)
            return {rch.reach: self.getSimData(riv.river, rch.reach, rs) for rch in riv.reaches}
        elif rs is None:
            riv = self.river(river)
            rch = self.reach(river, reach)
            return {xs.rs: self.getSimData(riv.river, rch.reach, xs.rs) for xs in rch.xses}
        else:
            riv = self.river(river)
            rch = riv.reach(reach)
            xs = rch.xs(rs)
            sd = self.ras.GetVelDist(xs.riverID, xs.reachID, xs.xsID, 0, prof)
            # sd: (6x args, (left station), (right station), (conv perc), (area), (wetted perimeter), (flow),
            # (depth), (velocity))
            velocity = sd[-1][0]
            depth = sd[-2][0]
            flow = sd[-3][0]
            etc = {
                "area": sd[-4][0],
                "wp": sd[-5][0]
            }
            return SimData(velocity, depth, flow, etc)

class SimData(object):
    """
    Simulation data for a cross section.  "etc" is a dictionary of less relevant data.
    """
    def __init__(self, velocity, maxDepth, flow, etc):
        self.velocity = velocity
        self.maxDepth = maxDepth
        self.flow = flow
        self.etc = etc

class XS(object):
    """
    A cross section.
    """
    def __init__(self, ras, river, reach, rs):
        """
        :param ras: RasObject (from wrapper)
        :param river: river (string)
        :param reach: string
        :param rs: river station (string)
        """
        self.ras = ras
        self.river = river
        self.reach = reach
        self.rs = rs
        self.riverID = getRiverID(self.ras, self.river)
        self.reachID = getReachID(self.ras, self.river, self.reach)
        self.xsID = getXSID(self.ras, self.river, self.reach, self.rs)

    def setAllManning(self, ns):
        """
        Set left, main, right ns.
        :param ns: [left, main, right] Manning's n
        """
        self.ras.SetMannLCR(self.river, self.reach,  self.rs, ns[0], ns[1], ns[2])

    def setMainChannelManning(self, n):
        """
        Set the main channel n.  For now, set left/right ns to be the same; in future, keep them as-is.
        :param n: main channel n.
        """
        self.ras.SetMannLCR(self.river, self.reach, self.rs, n, n, n)

    def editSteadyFlow(self):
        self.ras.EditSteadyFlow()

    def setSteadyFlow(self, flows, wait = False):
        self.ras.setSteadyFlow(self.river, self.reach, self.rs, flows, wait)


def getRiverID(ras, river):
    """
    Find river ID from river name
    :param ras: RasObject
    :param river: river name
    :return: river id
    """
    rivers = ras.GetRivers()[1]
    for ix, riv in enumerate(rivers):
        if riv.startswith(river): # starts with because the names often have a bunch of extra spaces
            return ix + 1
    return False

def getReachID(ras, river, reach):
    reaches = ras.GetReaches(getRiverID(ras, river))[2]
    for ix, rch in enumerate(reaches):
        if rch.startswith(reach):
            return ix + 1
    return False

def getXSID(ras, river, reach, xs):
    """
    Count from the top, not the bottom. <not currently>
    """
    xses = [xs.strip() for xs in ras.GetNodes(getRiverID(ras, river), getReachID(ras, river, reach))[3]]
    for (ix, rs) in enumerate(xses):
        if rs.startswith(xs):
            return ix + 1
    return False

class Reach(object):
    """
    A reach.
    """
    def __init__(self, ras, river, reach):
        self.ras = ras
        self.river = river
        self.reach = reach
        self.riverID = getRiverID(self.ras, self.river)
        self.reachID = getReachID(self.ras, self.river, self.reach)
        self.xses = [XS(self.ras, self.river, self.reach, rs) for rs in self.getCrossSections()]

    def getCrossSections(self):
        return [xs.strip() for xs in self.ras.GetNodes(self.riverID, self.reachID)[3]]

    def xs(self, rs):
        return [xs for xs in self.xses if xs.rs == rs][0]

    def xsAt(self, rs):
        return self.xs(rs)

class River(object):
    """
    A river.
    """
    def __init__(self, ras, river):
        self.ras = ras
        self.river = river
        self.riverID = getRiverID(self.ras, self.river)
        self.reaches = [Reach(self.ras, self.river, reach) for reach in self.getReaches()]

    def getReaches(self):
        return [r.strip() for r in self.ras.GetReaches(self.riverID)[2]]

    def reach(self, reach):
        return [rch for rch in self.reaches if rch.reach == reach][0]




"""
The API class.  No new functionality, but a place to gather together all relevant functions.
"""

"""
The following classes and methods are required (note: class names don't need to be as listed here,
but methods and properties do)

rasObj - containing a project with a specified geometry file and plan etc
rasObj.reach(river, reach, geom = None) - returning the relevant reach.  If geom (optional) is specified,
    use that geometry file (specified as an ID, if string, or file number, if integer); otherwise, use the currently
    active geometry, if any, or throw an error if no geometry is active
rasObj.openProject(projectPath) - open the relevant project
rasObj.setPlan(plan) - set the plan to that one (by ID or file number)
rasObj.setGeom(geom) - set the plan geometry file/active geometry file (by ID or file number)
rasObj.computeSteady, computeUnsteady(plan = None) - run the plan (optional) or, if plan is None, the currently active plan
rasObj.exit() - exit HEC-RAS
rasObj.newPlan(planId) - create a new plan with the given ID and make it the active plan
rasObj.set<Steady/Unsteady>Flow(flow) - set the steady/unsteady flow file (by ID or file number)
rasObj.getSimData(river = None, reach = None, rs = None): get the simulation flow data for a given river, reach, and river station.
    If rs is none, return a dictionary of all the river stations' data.  Likewise river and reach (nested dictionaries).
    Simulation flow data should be as a simdata class.

reach - containing a reach in the geometry file
reach.xses - return a list of cross sections

simdata - simulation data for a given cross section
properties: velocity, maxDepth, flow, and other data

xs - a cross section
xs.setAllManning(ns): set all Manning's ns to a list of ns given
xs.setMainChannelManning(n): set the main channel n to the given n
"""

from pyrasfile import profileWriter as pw

class OpsAPI(object):
    # Running and general operation
    def __init__(self, rasObj):
        self.ras = rasObj
    def openProject(self, projectPath):
        self.ras.openProject(projectPath)
    def setPlan(self, plan):
        self.ras.setPlan(plan)
    def setGeometry(self, geom):
        self.ras.setGeom(geom)
    def compute(self, steady = True, plan = None, wait = True):
        if steady:
            self.ras.computeSteady(plan)
        else:
            self.ras.computeUnsteady(plan)
        if wait:
            while not self.ras.computeIsComplete():
                pass
    def exit(self):
        self.ras.exit()
    def newPlan(self, planId):
        self.ras.newPlan(planId)
    def setFlow(self, flow, steady = True):
        if steady:
            self.ras.setSteadyFlow(flow)
        else:
            self.ras.setUnsteadyFlow(flow)

def nestedDictMap(map, func):
    # Apply func to all the values at the bottom level in map
    if map.__class__ == {}.__class__:
        for key in map:
            map[key] = nestedDictMap(map[key], func)
        return map
    else:
        return func(map)

class DataAPI(object):
    # Data retrieval
    def __init__(self, rasObj):
        self.ras = rasObj
    def allFlow(self, river = None, reach = None, rs = None, nprofs = 1):
        if nprofs == 1:
            return self.ras.getSimData(river, reach, rs)
        else:
            return {i: self.ras.getSimData(river, reach, rs, prof = i) for i in range(1, nprofs + 1)}
    def getSingleDatum(self, func, river, reach, rs, nprofs = 1):
        # Get a single datum (e.g. velocity, stage), regardless of level of nesting
        # func: function to extract relevant value from data class (e.g. lambda x: x.velocity)
        result = self.allFlow(river, reach, rs, nprofs)
        return nestedDictMap(result, func)
        # if not (rs is None): # just the one station
        #     return func(result)
        # elif not (reach is None): # just the one reach
        #     for rs, dat in result.items():
        #         dat = func(dat)
        #         result[rs] = dat
        #     return result
        # elif not (river is None): # just the one river
        #     for rch, rdat in result.items():
        #         for rs, dat in rdat.items():
        #             rdat[rs] = func(dat)
        #     return result
        # else: # all rivers
        #     for riv, ridat in result.items():
        #         for rch, rdat in ridat.items():
        #             for rs, dat in rdat.items():
        #                 rdat[rs] = func(dat)
        #     return result
    def velocity(self, river = None, reach = None, rs = None, nprofs = 1):
        return self.getSingleDatum(lambda x: x.velocity, river, reach, rs, nprofs)
    def stage(self, river = None, reach = None, rs = None, nprofs = 1):
        return self.getSingleDatum(lambda x: x.maxDepth, river, reach, rs, nprofs)
    # Below not strictly needed for raspy-cal
    # def ratingCurve
    # def n

class ParamsAPI(object):
    # Setting parameters
    def __init__(self, rasObj):
        self.ras = rasObj
    # def newSteadyFlow
    # def newUnsteadyFlow
    # def modifySteadyFlow
    # def modifyUnsteadyFlow
    # While the above methods will be useful, they are not critical to initial raspy-cal functionality
    def modifyN(self, manning, river, reach, geom = None):
        """
        Specify the Manning's n for the relevant geometry.
        :param manning: specifies Manning's n.  If it is a list of doubles, then this will set the main channel ns for
        the entire reach.  A list of lists of doubles will set that many ns at each point for the entire reach.
        Alternatively, a dictionary of <river station>:<n> can be used; if there is one n it will set the main channel,
        and if there is a list it will set all of them.  Note that Manning's equation seems to not be very sensitive to
        non-main channel n, in general. If the argument is just a number, it will set all main channel n to that number.
        Note that Manning's equation seems to not be very sensitive to non-main channel n, in general.

        Note that for now the list functionality will instead take a list of 3 to set left, main channel, right ns.
        :param river: name of the river.
        :param reach: name of the reach
        :param geom: name (if string) or number (if integer) of the geometry file.  If None, it will use the currently active geometry, if any.
        """
        if manning.__class__ == [].__class__ and len(manning) > 0:
            # List
            i = 0
            for xs in self.ras.reach(river = river, reach = reach, geom = geom).xses:
                if manning[i].__class__ == [].__class__:
                    # List of lists
                    xs.setAllManning(manning[i])
                else:
                    # List of individual ns
                    xs.setMainChannelManning(manning[i])
                i += 1
        elif manning.__class__ == {}.__class__:
            # Dictionary
            rch = self.ras.reach(river = river, reach = reach, geom = geom)
            for rs in manning:
                xs = rch.xsAt(rs)
                val = manning[rs]
                if val.__class__ == [].__class__:
                    xs.setAllManning(val)
                else:
                    xs.setMainChannelManning(val)
        elif manning.__class__ == 0.1.__class__:
            # Single value
            for xs in self.ras.reach(river, reach, geom).xses:
                xs.setMainChannelManning(manning)
        else:
            raise (TypeError("Manning must be a list, dictionary or float"))

    def editSteadyFlows(self):
        self.ras.editSteadyFlow()

    def setSteadyFlows(self, river, reach, rs, flows, slope = 0.001, fileN = "01", hecVer = "5.0.7"):
        # Slope is used for boundary conditions - normal depth
        # Bug: this doesn't take effect until Ras is closed and re-opened.
        rs = self.ras.reach(river, reach).xses[0].rs if rs is None else rs
        header = pw.mkFlowHeader(river, reach, rs)
        otherHeaders = []
        boundKeys = []
        for riv in self.ras.rivers:
            for rch in riv.reaches:
                topXs = rch.xses[0]
                otherHeaders.append(pw.mkFlowHeader(riv.river, rch.reach, topXs.rs))
                boundKeys.append("%s,%s" % (riv.river, rch.reach))
        count = len(flows)
        pdata = {h: [1] * count for h in otherHeaders}  # Meaningless data so HEC-RAS doesn't complain
        pdata[header] = flows
        bound = lambda pn, q: pw.mkBoundaryData("Normal Depth", "Normal Depth", slope, slope)
        bounds = {h: bound for h in boundKeys}
        flowFile = pw.buildFile(count, pdata, bounds, title="Flow" + fileN, ver=hecVer)
        projPath = self.ras.currentProject()
        basePath = "\\".join(projPath.split("\\")[:-1])
        projName = projPath.split("\\")[-1][:-4]  # without the .prj
        flowPath = "%s\\%s.f%s" % (basePath, projName, fileN)
        with open(flowPath, "w") as f:
            f.write(flowFile)
        # Save, quit, and reopen to make it use the new flow data
        self.ras.save()
        self.ras.quit()
        self.ras.openProject(projPath)



class API(object):
    def __init__(self, rasObj):
        self.ras = rasObj
        self.ops = OpsAPI(rasObj)
        self.data = DataAPI(rasObj)
        self.params = ParamsAPI(rasObj)


"""
General HEC-RAS operation: opening, closing, running, etc.
"""

PATH_SEP = "\\"

# ras, in all cases, is a Model object from the ras module

from ras import Model

def openProject(model, projectName, projectDir, basePath):
    projectPath = basePath + PATH_SEP + projectDir + PATH_SEP + projectName + ".prj"
    model.openProject(projectPath)
    return model    

def setPlan